"""
Package Purpose: registry of runs, including metadata, params, and metrics.
See also: https://gitlab.com/Sevans7/plasmacalcs/-/issues/44
"""

from .metadata_haver import MetadataHaver, pc_uuid_here, UuidHaver
from .register import Registry, Registerable
from .registry_tools import (
    registry_schema_sql,
    registry_location_default, registry_location, registry_name_default,
)
