"""
Package Purpose: misc. science tools
Like all tools, these should be portable, i.e. not specific to PlasmaCalcs.
These generally solve a very particular science problem (like ray draping),
    rather than doing more generic science-related computations (like curve fitting).
"""
from .ray_draping import (
    RayDrapeScene, RayDrapeAnchor, RayDrapeFOV, RayDrapingCore, RayDraping,
)
